package Bitmap;

import java.awt.*;
import java.io.*;
import java.util.*;
import LinearAlgebra.*;
import Paintable.*;

public class BitmapClass extends Object implements PaintableInterface,Loadable {
    final static int INITIAL_XSIZE = 20;
    final static int INITIAL_YSIZE = INITIAL_XSIZE;
    private int xSize;
    private int ySize;
    private int maxXSize;
    private int maxYSize;
    private int [][] template;
    
    public BitmapClass() {
	maxXSize = 0;
	maxYSize = 0;
	SetSizes(INITIAL_XSIZE,INITIAL_YSIZE);
    }

    public BitmapClass(BitmapClass bitmap) {
	maxXSize = bitmap.maxXSize;
	maxYSize = bitmap.maxYSize;
	xSize = bitmap.xSize;
	ySize = bitmap.ySize;
	template = bitmap.template;
    }

    public BitmapClass(int in_xSize,int in_ySize) {
	maxXSize = 0;
	maxYSize = 0;
	SetSizes(in_xSize,in_ySize);
    }

    public BitmapClass Clone() {
	BitmapClass clone = new BitmapClass(xSize,ySize);
	for(int i=0;i<xSize;i++)
	    for(int j=0;j<ySize;j++)
		clone.SetPixel(i,j,GetPixel(i,j));
	return clone;
    }

    final public void Normalize(int newXSize,int newYSize) {
	BitmapClass oldBitmap = Clone();
	double factorX = (double)newXSize/xSize;
	double factorY = (double)newYSize/ySize;
	Reset();
	SetSizes(newXSize,newYSize);
	double grayLevel;
	int startX,startY,endX,endY;
	double factorStartX,factorStartY,factorEndX,factorEndY;
	for(int i=0;i<xSize;i++) {
	    startX = (int)(i/factorX + 0.5/xSize);
	    endX = (int)((i+1)/factorX - 0.5/xSize);
	    factorStartX = i - startX*factorX;
	    factorEndX = (endX+1)*factorX - (i+1);
	    for(int j=0;j<ySize;j++) {
		startY = (int)(j/factorY + 0.5/ySize);
		endY = (int)((j+1)/factorY - 0.5/ySize);
		factorStartY = j - startY*factorY;
		factorEndY = (endY+1)*factorY - (j+1);
		grayLevel = 0;
		for(int i1=startX;i1<=endX;i1++)
		    for(int j1=startY;j1<=endY;j1++)
			grayLevel += oldBitmap.GetPixelExtended(i1,j1) * factorX * factorY;
		for(int i1=startX;i1<=endX;i1++) {
		    grayLevel -= oldBitmap.GetPixelExtended(i1,startY) * factorX * factorStartY;
 		    grayLevel -= oldBitmap.GetPixelExtended(i1,endY) * factorX * factorEndY;
		}
 		for(int j1=startY;j1<=endY;j1++) {
 		    grayLevel -= oldBitmap.GetPixelExtended(startX,j1) * factorStartX * factorY;
		    grayLevel -= oldBitmap.GetPixelExtended(endX,j1) * factorEndX * factorY;
		}
 		grayLevel += oldBitmap.GetPixelExtended(startX,startY) * factorStartX * factorStartY;
 		grayLevel += oldBitmap.GetPixelExtended(startX,endY) * factorStartX * factorEndY;
 		grayLevel += oldBitmap.GetPixelExtended(endX,startY) * factorEndX * factorStartY;
 		grayLevel += oldBitmap.GetPixelExtended(endX,endY) * factorEndX * factorEndY;
//		System.out.println(grayLevel);
		if (grayLevel >= 0.5)
		    SetPixel(i,j);

	    }
	}	
    }
	
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                    interface Loadable BEGIN
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    final public void Load(File f) throws FileNotFoundException, IOException {
	FileInputStream fin = new FileInputStream(f);
	DataInputStream din = new DataInputStream(fin);
	Load(din);
	din.close();
	fin.close(); 
    }

    final public void Load(File f,Debug.Debug d) throws FileNotFoundException, IOException {
	FileInputStream fin = new FileInputStream(f);
	DataInputStream din = new DataInputStream(fin);
	Load(din);
	din.close();
	fin.close(); 
    }

    final public void Load(DataInputStream din,Debug.Debug d) throws IOException {
	Load(din);
    }

    final public void Load(DataInputStream din) throws IOException {
	int rowNum = 0;
	int colNum = 0;
	try {
	    String line;
	    StringTokenizer t;               
	    int xS = 0,yS = 0;               
	    while(true) {
		line = new String(din.readLine());                    
		t = new StringTokenizer(line);
		try {
		    xS = (new Long(t.nextToken())).intValue();
		    yS = (new Long(t.nextToken())).intValue();
		    break;
		}
		catch(RuntimeException e) {
		}
	    }
//  	    System.out.println(xS + " " + yS);                           
	    if (xS < 2 || yS < 2)
		throw new NoSuchElementException();
	    SetSizes(xS,yS);                                             
	    for(int j=GetYSize() - 1;j>=0;j--) {                         
		rowNum++;
		for(int i=0;i<GetXSize();i++) {                          
		    colNum++;
		    if (!t.hasMoreTokens()) {                            
			line = new String(din.readLine());               
			t = new StringTokenizer(line);                   
		    }                                                    
		    SetPixel(i,j,(new Long(t.nextToken())).intValue());  
		}                                                        
		colNum = 0;
	    }                                                            
	}
	catch (NoSuchElementException exception3) {
	    System.err.println("Wrong portable bitmap format (NoSuchElementException)");
	    System.err.println("Row number = " + rowNum);
	    System.err.println("Column number = " + colNum);
	    Reset();
	}
	catch (NumberFormatException exception4) {
	    System.err.println("Wrong portable bitmap format (NumberFormatException)");
	    Reset();
	}
    }

//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                    interface Loadable END
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////

    final public void SetSizes(int in_xSize,int in_ySize) {
	xSize = in_xSize;
	ySize = in_ySize;	
	if (xSize > maxXSize) {
	    int [][] oldTemplate = template;
	    template = new int[xSize][];
	    if (ySize > maxYSize) {
		for(int i=0;i<maxXSize;i++) {
		    template[i] = new int[ySize];
		    for(int j=0;j<maxYSize;j++) {
			template[i][j] = oldTemplate[i][j];
		    }
		    for(int j=maxYSize;j<ySize;j++) {
			template[i][j] = 0;
		    }
		}
		for(int i=maxXSize;i<xSize;i++) {
		    template[i] = new int[ySize];
		    for(int j=0;j<ySize;j++) {
			template[i][j] = 0;
		    }
		}			
		maxYSize = ySize;
	    }
	    else {
		for(int i=0;i<maxXSize;i++) {
		    template[i] = oldTemplate[i];
		}
		for(int i=maxXSize;i<xSize;i++) {
		    template[i] = new int[maxYSize];
		    for(int j=0;j<maxYSize;j++) {
			template[i][j] = 0;
		    }
		}
	    }
	    maxXSize = xSize;
	}
	else if (ySize > maxYSize) {
	    for(int i=0;i<maxXSize;i++) {
		int [] oldTemplateI = template[i];
		template[i] = new int[ySize];
		for(int j=0;j<maxYSize;j++) {
		    template[i][j] = oldTemplateI[j];
		}
		for(int j=maxYSize;j<ySize;j++) {
		    template[i][j] = 0;
		}
	    }
	    maxYSize = ySize;
	}
	for(int i=xSize;i<maxXSize;i++)
	    for(int j=0;j<maxYSize;j++)
		template[i][j] = 0;
	for(int j=ySize;j<maxYSize;j++)
	    for(int i=0;i<xSize;i++)
		template[i][j] = 0;
    }

    final public int GetXSize() {
	return xSize;
    }
    
    final public int GetYSize() {
	return ySize;
    }
    
    final public int GetPixel(int i,int j) {
	try {
	    return template[i][j];
	}
	catch(ArrayIndexOutOfBoundsException e) {
	    String s = new String();
	    s += "i = " + i + "\n";
	    s += "j = " + j + "\n";
	    s += "xSize = " + xSize + "\n";
	    s += "ySize = " + ySize + "\n";
	    throw new ArrayIndexOutOfBoundsException(s);
	}   
    }
    
    //returns 0 if invalid pixels are asked
    final public int GetPixelExtended(int i,int j) {
	try {
	    return template[i][j];
	}
	catch(ArrayIndexOutOfBoundsException e) {
	    return 0;
	}   
    }
    
    final public boolean GetPixel(int i,int j,int threshold) {
	return GetPixel(i,j) > threshold;
    }

    //returns false if invalid pixels are asked
    final public boolean GetPixelExtended(int i,int j,int threshold) {
	return GetPixelExtended(i,j) > threshold;
    }
    
    final public void SetPixel(int i,int j,int value) {
	try {
	    template[i][j] = value;
	}
	catch(ArrayIndexOutOfBoundsException e) {
	    String s = new String();
	    s += "i = " + i + "\n";
	    s += "j = " + j + "\n";
	    s += "xSize = " + xSize + "\n";
	    s += "ySize = " + ySize + "\n";
	    throw new ArrayIndexOutOfBoundsException(s);
	}   
    }
    
    final public void SetPixel(int i,int j) {
	SetPixel(i,j,1);
    }

    final public void ResetPixel(int i,int j) {
	SetPixel(i,j,0);
    }

    final public void Reset() {
	for(int i=0;i<xSize;i++)
	    for(int j=0;j<ySize;j++)
		ResetPixel(i,j);
    }

    // sample will be normalized so that the its bottom-left
    // corner is at (0,0), and 1 unit is 1 pixel on both axes,
    // and each sample point represents the middle of the pixel.
    // Each pixel is thresholded at 0
    final public Sample ConvertToSample() {
	return ConvertToSample(0);
    }
    
    final public Sample ConvertToSample(int threshold) {
	Sample sample = new Sample();
	for(int i=0;i<xSize;i++) {
	    for(int j=0;j<ySize;j++) {
		if (GetPixel(i,j,threshold))
		    sample.AddPoint(ConvertPixelToVektorDD(i,j));
	    }
	}
	return sample;
    }
    
    final public Sample ConvertToSampleLabeled() {
	Sample sample = new Sample();
	for(int i=0;i<xSize;i++) {
	    for(int j=0;j<ySize;j++) {
		sample.AddPoint(ConvertPixelToVektorDDLabeled(i,j));
	    }
	}
	return sample;
    }
    
    final protected VektorDD ConvertPixelToVektorDD(int i,int j) {
	double [] coords = {i + 0.5,j + 0.5};
	return new VektorDD(coords);
    }

    final protected VektorDDLabeled ConvertPixelToVektorDDLabeled(int i,int j) {
	double [] coords = {i + 0.5,j + 0.5};
	return new VektorDDLabeled(coords,template[i][j]);
    }

    final public double Dist1(BitmapClass bitmap) {
	try {
	    int dist = 0;
	    for(int i=0;i<xSize;i++)
		for(int j=0;j<ySize;j++)
		    dist += Math.abs(template[i][j] - bitmap.template[i][j]);
	    return dist;
	}
	catch (ArrayIndexOutOfBoundsException e) {
	    throw new ArrayIndexOutOfBoundsException("Different bitmap sizes:\nOwner: " + xSize + " x " + ySize + 
						     "parameter: " + bitmap.xSize + " x " + bitmap.ySize + "\n");
	}
    }
	    

    final public double GetEastBorder(ProjectionPlane projectionPlane) {
	return GetXSize();
    }

    final public double GetWestBorder(ProjectionPlane projectionPlane) {
	return 0;
    }

    final public double GetSouthBorder(ProjectionPlane projectionPlane) {
	return 0;
    }
    
    final public double GetNorthBorder(ProjectionPlane projectionPlane) {
	return GetYSize();
    }

    final public void Paint(Graphics g,DataCanvas canvas,int pixelSize,String type) {
	Sample sample = ConvertToSample();
	type = VektorObject.SQUARE_POINT_TYPE;
	pixelSize = (int)Math.min(canvas.ConvertWidth(1),canvas.ConvertHeight(1));
	sample.Paint(g,canvas,pixelSize,type);
    }

    final public void PrintToPostScript(DataPostScriptDocument ps,double pixelSize,String type,String title) {
	Sample sample = ConvertToSample();
	type = VektorObject.SQUARE_POINT_TYPE;
	pixelSize = ps.ConvertWidth(1)/ps.GetWidthUnit();
	sample.PrintToPostScript(ps,pixelSize,type,title);
    }

    final public void Save(String fileName) {
	try {
	    FileOutputStream fOut = new FileOutputStream(fileName);
	    PrintStream pOut = new PrintStream(fOut);
	    pOut.println(xSize + " " + ySize);
	    int n = 0;
	    int maxN = 40;
	    if (GetXSize() < 40)
		maxN = GetXSize();
	    for(int j=GetYSize() - 1;j>=0;j--) {                         
		for(int i=0;i<GetXSize();i++) {                          
		    if (n >= maxN) {
			pOut.println();
			n = 0;
		    }
		    n++;
		    pOut.print(GetPixel(i,j) + " ");
		}                                                        
	    }                                                            
	    pOut.close();
	    fOut.close();
	} catch(IOException e) {
	    System.out.println("Can't open file " + fileName);
	}
    }

    final public String toString() {
	String s = new String();
// 	for(int i=0;i<xSize;i++) {
// 	    for(int j=0;j<ySize;j++) {
// 		s += "pixel = (" + i + "," + j + "),label = " + GetPixel(i,j) + "\n";
// 	    }
// 	} 	
	for(int j=GetYSize() - 1;j>=0;j--) {
	    for(int i=0;i<GetXSize();i++) {
		s += GetPixel(i,j) + " ";
	    }
	    s += "\n";
	} 	
	return s;
    }
}



