package LinearAlgebra;

import java.awt.*;
import java.io.*;
import java.lang.*;
import java.util.*;
import Paintable.*;

public class Sample extends Object implements PaintableInterface {
    protected Vector points;

    public Sample() {
	InitializeSampleParameters(100);
    }

    public Sample(int initSize) {
	InitializeSampleParameters(initSize);
    }

    final private void InitializeSampleParameters(int initSize) {
	points = new Vector(initSize);
    }

    public Sample Clone() {
	Sample sample = new Sample(GetSize());
	for(int i=0;i<GetSize();i++)
	    sample.AddPoint(GetPointAt(i).Clone());
	return sample;
    }

    public Sample ShallowClone() {
	Sample sample = new Sample(GetSize());
	for(int i=0;i<GetSize();i++)
	    sample.AddPoint(GetPointAt(i));
	return sample;
    }

    public Sample DefaultClone() {   
	return new Sample();
    }    

    final public Vektor GetPointAt(int i) {
	try {
	    return (Vektor)(points.elementAt(i));
	}
	catch (ArrayIndexOutOfBoundsException e) {
	    throw new ArrayIndexOutOfBoundsException("Myerror: i = " + i + " size = " + GetSize() + "\n");
	}
    }

    final public int FindPoint(Vektor vektor) {
	return points.indexOf(vektor);
    }

    final public int GetSize() {
        return points.size();
    }

    public String toString() {
	String string = "Size = " + GetSize() + "\n";
	for(int i=0;i<GetSize();i++)
	    string += i + ": " + GetPointAt(i).toString() + "\n";
	return string;
    }

    public void Reset() {
	points.setSize(0);
    }

    final public void Add(Sample sample) {
	for(int i=0;i<sample.GetSize();i++)
	    AddPoint(sample.GetPointAt(i));
    }

    final public void AddPoint(Vektor vektor) {
	InsertPointAt(vektor,GetSize());
    }

    public void InsertPointAt(Vektor vektor,int index) {
	try {
	    points.insertElementAt(vektor,index);
	}
	catch (ArrayIndexOutOfBoundsException e) {
	    throw new ArrayIndexOutOfBoundsException("Myerror: index = " + index + " size = " + GetSize() + "\n");
	}
    } 

    public void DeletePointAt(int index) {
	try {
	    points.removeElementAt(index);
	}
	catch (ArrayIndexOutOfBoundsException e) {
	    throw new ArrayIndexOutOfBoundsException("Myerror: index = " + index + " size = " + GetSize() + "\n");
	}
    }

    // Batch deleting
    public void DeletePoints(boolean [] toBeDeleted) {
	int j = 0;
	for(int i=0;i<GetSize();i++) {
	    if (!toBeDeleted[i]) {
		points.setElementAt(points.elementAt(i),j);
		j++;
	    }
	}
	points.setSize(j);
    }

    public void UpdatePointAt(Vektor vektor,int i) {
	GetPointAt(i).Update(vektor);
    }

    public void SetPointAt(Vektor vektor,int index) {
	try {
	    points.setElementAt(vektor,index);
	}
	catch (ArrayIndexOutOfBoundsException e) {
	    throw new ArrayIndexOutOfBoundsException("Myerror: index = " + index + " size = " + GetSize() + "\n");
	}
    }

    // Creating smaller random sample with min(maxpoints,size) points 
    final public Sample RandomSample(int maxpoints) {
	Sample random_sample = DefaultClone();
	for(int i=0;i<GetSize();i++)
	    if (Math.random() < (double)(maxpoints - random_sample.GetSize())/(GetSize() - i) )
		random_sample.AddPoint(GetPointAt(i));
	return random_sample;
    }

    final public Sample GetProjectionResiduals(Line line) {
	Sample residuals = DefaultClone();
	for(int i=0;i<GetSize();i++) {
	    Vektor vektor = GetPointAt(i);
	    Vektor projection = vektor.Project(line);
	    residuals.AddPoint(vektor.Sub(projection));
	}
	return residuals;
    }

    public void Save(String fileName) {
	try {
	    FileOutputStream fOut = new FileOutputStream(fileName);
	    PrintStream pOut = new PrintStream(fOut);
	    for(int i=0;i<GetSize();i++) {
		GetPointAt(i).Save(pOut);
		pOut.println();
	    }
	    pOut.close();
	    fOut.close();
	} catch(IOException e) {
	    System.out.println("Can't open file " + fileName);
	}
    }
    
    public String SaveToString() {
	String str = new String();
	for(int i=0;i<GetSize();i++) {
	    str += GetPointAt(i).SaveToString() + "\n";
	}
	return str;
    }
    
    public void AddEqual(Vektor vektor) {
	for(int i=0;i<GetSize();i++)
	    GetPointAt(i).AddEqual(vektor);
    }
	
    public void SubEqual(Vektor vektor) {
	for(int i=0;i<GetSize();i++)
	    GetPointAt(i).SubEqual(vektor);
    }

    public void MulEqual(double d) {
	for(int i=0;i<GetSize();i++)
	    GetPointAt(i).MulEqual(d);
    }
	
    public void DivEqual(double d) {
	for(int i=0;i<GetSize();i++)
	    GetPointAt(i).DivEqual(d);
    }
	
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                    interface PaintableInterface BEGIN
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
    public void Paint(Graphics g,DataCanvas canvas,int pixelSize,String type) {
	for(int i=0;i<GetSize();i++)
	    ((VektorObject)GetPointAt(i)).Paint(g,canvas,pixelSize,type);
    }

    public void PrintToPostScript(DataPostScriptDocument ps,double pixelSize,String type,String title) {
	if (GetSize() > 0) {
	    ps.GSave();
	    double [] dashPattern = new double[0];
	    ps.SetDash(dashPattern,0);
	    boolean changingIntensity = false;
	    int s = Math.min(2000,GetSize());
	    for(int i=0;i<s;i++) {
		ps.NewLine();
		ps.Push(ps.ConvertXCoordinate(GetPointAt(i).GetCoordX(ps.projectionPlane)));
		ps.Push(ps.ConvertYCoordinate(GetPointAt(i).GetCoordY(ps.projectionPlane)));
		try {
		    PaintableWithChangingIntensity point = (PaintableWithChangingIntensity)GetPointAt(i);
		    ps.Push(DataPostScriptDocument.ConvertRGBColor(
			point.ChangeRGBComponentIntensity(ps.GetColor().getRed())));
		    ps.Push(DataPostScriptDocument.ConvertRGBColor(
			point.ChangeRGBComponentIntensity(ps.GetColor().getGreen())));
		    ps.Push(DataPostScriptDocument.ConvertRGBColor(
			point.ChangeRGBComponentIntensity(ps.GetColor().getBlue())));
		    changingIntensity= true;
		}
		catch(ClassCastException e) {
		}
	    }
	    ps.NewLine();
	    ps.BeginFor(0,1,s);
	    {
		ps.Pop();
		if (changingIntensity)
		    ps.SetRGBColor();
		ps.DrawPoint(pixelSize,type);
	    }
	    ps.EndFor();
	    if (!title.equals("")) {
		ps.Push(ps.GetNextLineCenterX());
		ps.Push(ps.GetNextLineCenterY());
		ps.DrawPoint(pixelSize,type);
		if (ps.BlackAndWhite())
		    ps.SetGray(0); //black
		else 
		    ps.SetRGBColor(ps.GetColor());
		ps.PrintText(title,ps.GetNextLineTextX(),ps.GetNextLineTextY());
		ps.NextLine();
	    }
	    ps.GRestore();
	}    
    }

    final public double GetEastBorder(ProjectionPlane projectionPlane) {
	if (GetSize() == 0)
	    return 0.0;
	double d = GetPointAt(0).GetCoordX(projectionPlane);
	for(int i=0;i<GetSize();i++) {
	    if (d < GetPointAt(i).GetCoordX(projectionPlane)) {
		d = GetPointAt(i).GetCoordX(projectionPlane);
	    }
	}
	return d;
    }

    final public double GetWestBorder(ProjectionPlane projectionPlane) {
	if (GetSize() == 0)
	    return 0.0;
	double d = GetPointAt(0).GetCoordX(projectionPlane);
	for(int i=0;i<GetSize();i++) {
	    if (d > GetPointAt(i).GetCoordX(projectionPlane)) {
		d = GetPointAt(i).GetCoordX(projectionPlane);
	    }
	}
	return d;
    }

    final public double GetSouthBorder(ProjectionPlane projectionPlane) {
	if (GetSize() == 0)
	    return 0.0;
	double d = GetPointAt(0).GetCoordY(projectionPlane);
	for(int i=0;i<GetSize();i++) {
	    if (d > GetPointAt(i).GetCoordY(projectionPlane)) {
		d = GetPointAt(i).GetCoordY(projectionPlane);
	    }
	}
	return d;
    }

    final public double GetNorthBorder(ProjectionPlane projectionPlane) {
	if (GetSize() == 0)
	    return 0.0;
	double d = GetPointAt(0).GetCoordY(projectionPlane);
	for(int i=0;i<GetSize();i++) {
	    if (d < GetPointAt(i).GetCoordY(projectionPlane)) {
		d = GetPointAt(i).GetCoordY(projectionPlane);
	    }
	}
	return d;
    }
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
//                                    interface PaintableInterface END
//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
}



