package LinearAlgebra;

import java.awt.*;
import java.io.*;
import java.util.*;
import Paintable.*;

final public class VektorDDWeighted extends VektorDD implements PaintableWithChangingIntensity,Weighted {
    protected double weight;
    public static double maxWeight = 1.0; // for painting

    protected VektorDDWeighted() {
	super();
	weight = 1;
    }

    public VektorDDWeighted(int dimension,double in_weight) {
	super(dimension);
	weight = in_weight;
    }

    public VektorDDWeighted(double [] in_coords,double in_weight) {
	super(in_coords);
	weight = in_weight;
    }

    public VektorDDWeighted(Vektor vektor,double in_weight) {
	super(vektor);
	weight = in_weight;
    }

    public VektorDDWeighted(StringTokenizer t) throws NoSuchElementException, NumberFormatException {
	super(t.countTokens() - 1);
	Constructor(t);
	weight = (new Double(t.nextToken())).doubleValue();
    }

    final public Vektor Clone() {
	return new VektorDDWeighted(coords,weight);
    }

    final public Vektor DefaultClone() {
	return new VektorDDWeighted();
    }

    final public double GetWeight() {
	return weight;
    }

    final public void SetWeight(double in_weight) {
	weight = in_weight;
    }

    final public void Paint(Graphics g,DataCanvas canvas,int pixelSize,String type) {
	Color oldColor = g.getColor();
	g.setColor(new Color(ChangeRGBComponentIntensity(oldColor.getRed()),
			     ChangeRGBComponentIntensity(oldColor.getGreen()),
			     ChangeRGBComponentIntensity(oldColor.getBlue())));
	super.Paint(g,canvas,pixelSize,type);
	g.setColor(oldColor);
    }

    final public int ChangeRGBComponentIntensity(int c) {
	return (int)(255 - (255 - c)*weight/maxWeight);
    }

    final public String toString() {
	return super.toString() + "w = " + Utilities.MyMath.RoundDouble(weight,4);
    }

    final public void Save(PrintStream pOut) {
	super.Save(pOut);
	pOut.print(" " + weight);	
    }

    final public String SaveToString() {
	return super.SaveToString() + " " + weight;	
    }
}



