package PrincipalCurve;

import java.awt.*;
import java.io.*;
import java.util.*;

import AWT.*;
import Bitmap.*;
import Debug.*;
import Internet.*;
import LinearAlgebra.*;
import Paintable.*;

final public class SkeletonizationGUI extends Frame implements Recallable {
    final String logDirectory = "PrincipalCurveLogs/";
    final String outputDirectory = "PrincipalCurveOutputs/";
    //    Internet.UpLoadFile logFile = new Internet.UpLoadFile(logDirectory + "log");
    Internet.UpLoadFile logFile = null;
    String templateFileName = "image.bm";
    final String pgraphFileName = "pgraph.dta";
 
    //Dialogs
    FileDialog loadDialog = null;
    FileDialog saveDialog = null;
    DownLoadDialog downLoadDialog = null;
    int numOfEmails = -1;
    InitCurveDialog initCurveDialog = null;
    ParametersDialog parametersDialog = null;
   
    SkeletonizationCanvas canvas = new SkeletonizationCanvas();
    
    //Principal Curve Members
    Sample sample;
    Template characterTemplate;
    PrincipalCurveOfTemplate principalCurve;
    Loadable [] loadableData = new Loadable[2]; //{sample,principalCurve};

    // Menus
    MenuBar menuBar = new MenuBar();
    MenuItem separator = new MenuItem("-");

    Menu sampleMenu = new Menu("Template");
    Menu builtInSamplesMenu = new Menu("Built in templates");
    Menu [] digitsMenus;
    Menu [][] digitsMenus1;
    MenuItem [][][] digitsMenuItems;
    MenuItem loadMenuItem = new MenuItem("Load template...");
    MenuItem downLoadMenuItem = new MenuItem("Download template...");
    MenuItem saveMenuItem = new MenuItem("Save data...");
    MenuItem saveImageMenuItem = new MenuItem("Save image...");
    MenuItem quitMenuItem = new MenuItem("Quit");

    Menu principalCurveMenu = new Menu("Principal graph");
    MenuItem parametersMenuItem = new MenuItem("Parameters...");
    MenuItem initCurveMenuItem = new MenuItem("Initialize graph...");

    Menu appearanceMenu = new Menu("Appearance");
    MenuItem packMenuItem = new MenuItem("Pack");
    MenuItem colorsAndSizesMenuItem = new MenuItem("Colors & sizes...");
 
    Menu [] menuBarElements = {sampleMenu,
			       principalCurveMenu,
			       appearanceMenu
    };

    Menu [] menus = {sampleMenu,
		     principalCurveMenu,
		     appearanceMenu
    };

    MenuItem [][] menuItems = { {builtInSamplesMenu,
				 loadMenuItem,
				 downLoadMenuItem,
				 saveMenuItem,
				 saveImageMenuItem,
				 separator,
				 quitMenuItem},

				{parametersMenuItem,
				 initCurveMenuItem},

				{packMenuItem,
				 colorsAndSizesMenuItem}
    };

    private void InitializeMenuBar() {
	setMenuBar(menuBar);
	for(int i=0;i<menuBarElements.length;i++)
	    menuBar.add(menuBarElements[i]);
	for(int i=0;i<menus.length;i++) {
	    for(int j=0;j<menuItems[i].length;j++) {
		if (shouldAddComponent(menuItems[i][j])) {
		    menus[i].add(menuItems[i][j]);
		}
	    }
	}
	digitsMenus = new Menu[10];
	digitsMenus1 = new Menu[10][];
	digitsMenuItems = new MenuItem[10][][];
	for(int i=0;i<10;i++) {
	    digitsMenus[i] = new Menu(i + "'s");
	    builtInSamplesMenu.add(digitsMenus[i]);
	    digitsMenus1[i] = new Menu[10];
	    digitsMenuItems[i] = new MenuItem[10][];
	    for(int j=0;j<10;j++) {
		digitsMenus1[i][j] = new Menu(j + "0");
		digitsMenus[i].add(digitsMenus1[i][j]);
		digitsMenuItems[i][j] = new MenuItem[10];
		for(int k=0;k<10;k++) {
		    digitsMenuItems[i][j][k] = new MenuItem(k + "");
		    digitsMenus1[i][j].add(digitsMenuItems[i][j][k]);
		}
	    }
	}

    }

    // Buttons
    Button startButton = new Button("Start");
    Button stopButton = new Button("Stop");
    Button initButton = new Button("Initialize");
    Button fitAndSmoothButton = new Button("Fit & Smooth");
    Button restructureButton = new Button("Restructure");

    Button [] buttons = {startButton,
			 stopButton,
			 initButton,
			 fitAndSmoothButton,
			 restructureButton};

    // Text fields
    TextField diagnosisTextField = new TextField();

    TextField [] textFields = {diagnosisTextField};
    boolean [] textFieldEditables = {false};

    private void InitializeTextFields() {
	for(int i=0;i<textFields.length;i++) {
	    textFields[i].setEditable(textFieldEditables[i]);
	}
    }

    // Gridlayout panels
    Panel canvasPanel = new Panel();
    Panel algorithmManagingPanel = new Panel();
    Panel messagePanel = new Panel();

    Panel [] gridPanels = {canvasPanel,
			   algorithmManagingPanel,
			   messagePanel
    };


    Component [][] gridPanelComponents = { {canvas},

					   {startButton,
					    stopButton,
					    initButton,
					    fitAndSmoothButton,
					    restructureButton},

					   {diagnosisTextField}
    };

    
    int [][] gridPanelLayoutDimensions = { {1,1},
					   {0,1},
					   {1,0}
    };

    private void InitializeGridPanels() {
 	for(int i=0;i<gridPanels.length;i++)
	    InitializeGridPanel(i);
    }
    
    private void InitializeGridPanel(int i) {
	gridPanels[i].setLayout(new GridLayout(gridPanelLayoutDimensions[i][0],
					       gridPanelLayoutDimensions[i][1]));
	for(int j=0;j<gridPanelComponents[i].length;j++) {
	    if (shouldAddComponent(gridPanelComponents[i][j]))
		gridPanels[i].add(gridPanelComponents[i][j]);
	}
    }
    
    private void InitializeGridPanel(Panel panel) {
 	for(int i=0;i<gridPanels.length;i++) {
	    if (panel == gridPanels[i]) {
		InitializeGridPanel(i);
		return;
	    }
	}
    }

    // The main panel (borderlayout)
    Component [] mainPanels = {canvasPanel,
			       algorithmManagingPanel,
			       messagePanel};
    
    String [] mainPanelPositions = {"Center",
				    "East",
				    "South"};
				    
    private void InitializeMainPanel() {
	setLayout(new BorderLayout());
	for(int i=0;i<mainPanels.length;i++) {
	    if (shouldAddComponent(mainPanels[i])) {
		add(mainPanelPositions[i],mainPanels[i]);
	    }
	}
    }
		
    private void InitializePanels() {
	InitializeMainPanel();
	InitializeGridPanels();
    }
		
    // The components that should not appear if we are in applets
    Object [] nonAppletComponents = {loadMenuItem,saveMenuItem,saveImageMenuItem,initCurveMenuItem};
    boolean shouldAddComponent(Object component) {
	if (Utilities.Environment.inApplet)
	    for(int i=0;i<nonAppletComponents.length;i++)
		if (component == nonAppletComponents[i])
		    return false;
	return true;
    }

    // Samples to download
    final String URL_SAMPLE_DIRECTORY = Utilities.Environment.homeURL + "/research/pcurves/implementations/SkeletonizationTemplates/";
    long timeStart;
    int xPosition;
    int yPosition;
    
    PrincipalCurveAlgorithmThreadForSkeletonization algorithmThread;    

    boolean onlyInitialize;
    boolean onlyRestructure;
    boolean onlyFitAndSmooth;

    private void InitializeCanvas() {
	canvas.SetObjectToPaint(canvas.sample,sample);
	canvas.SetObjectToPaint(canvas.principalCurve,principalCurve);
	canvas.SetObjectToPaint(canvas.principalCurvePoints,principalCurve);
    }
	
    public SkeletonizationGUI() {
	Initialize();
    }

    long zeroTime = (new Date()).getTime();

    public void Initialize() {
        setTitle("Principal Graph Skeletonization");
  	if (logFile != null && Utilities.Environment.inApplet) {
	    try {
		logFile.SetAction(Internet.UpLoadFile.ACTION_OPEN);
		logFile.AppendString((new Date()).toString() + "\n");
	    }
	    catch(RuntimeException e) {
		logFile = null;
	    }
  	}	    
	xPosition = 0;
	yPosition = 0;

	InitializeTextFields();
	InitializePanels();
	InitializeCanvas();
	InitializeMenuBar();

	Debug.diagnosisTextComponent = diagnosisTextField;
	Globals.diagnosisTextArea = new TextArea(20,80);
	
	onlyInitialize = false;
	onlyRestructure = false;
	onlyFitAndSmooth = false;

	DownLoad(URL_SAMPLE_DIRECTORY + "Number0/n000.bm");
    }

    public boolean action(Event event, Object arg) {
	if (logFile != null && Utilities.Environment.inApplet) {
	    logFile.AppendString((((new Date()).getTime() - zeroTime)/1000.0) + ":\t" + arg.toString() + "\n");
	}	    
	Object target = event.target;

	// Load files
	if (target == loadMenuItem) { 
	    if (loadDialog == null) {
		loadDialog = new FileDialog(this,"Choose a data file");
		if (saveDialog == null)
		    loadDialog.setDirectory(Utilities.Environment.homeDirectory + "/Research/Samples/Nist/NumberBitmaps100/");
		else
		    loadDialog.setDirectory(saveDialog.getDirectory());		    
	    }
	    loadDialog.show();
	    if (loadDialog.getFile() != null) {
		Load(loadDialog.getDirectory(),loadDialog.getFile());
	    }		
	}
	// Download files
	else if (target == downLoadMenuItem) { 
	    if (downLoadDialog == null) {
		downLoadDialog = new DownLoadDialog(this,"Download from url (directory of your data files)",
						    Utilities.Environment.homeURL + 
						    "/research/pcurves/implementations/SkeletonizationTemplates/");
	    }
	    downLoadDialog.show();
	    DownLoad(downLoadDialog.getURL());
	}
	// Save files
	if (target == saveMenuItem) { 
	    if (saveDialog == null) {
		saveDialog = new FileDialog(this,"Choose a data file",FileDialog.SAVE);
		if (loadDialog == null)
		    saveDialog.setDirectory("/tmp/");
		else
		    saveDialog.setDirectory(loadDialog.getDirectory());		    
	    }
	    saveDialog.show();
	    if (saveDialog.getFile() != null) {
		if (!saveDialog.getFile().equals(""))
		    templateFileName = saveDialog.getFile();
		characterTemplate.Save(saveDialog.getDirectory() + templateFileName); 			    
		if (principalCurve.GetSize() > 0) {
		    SetOfCurves savePrincipalCurve = principalCurve.ConvertToCurves();
		    if (savePrincipalCurve != null) {
			savePrincipalCurve.Save(saveDialog.getDirectory() + pgraphFileName);
		    }
		}
	    }		
	}
	// Save image
	else if (target == saveImageMenuItem) { 
 	    canvas.SaveImage(this);
	}
	// Set algorithm parameters
	else if (target == parametersMenuItem) {
	    if (parametersDialog == null) {
		parametersDialog = new ParametersDialog(this);
	    }
	    parametersDialog.show();
	} 
	// Initialize curve
	else if (target == initCurveMenuItem) {
	    initCurveDialog = new InitCurveDialog(this,canvas);
	    initCurveDialog.show();
	} 
	// Start algorithm
	else if (target == startButton) {
	    algorithmThread.SetAction(PrincipalCurveAlgorithmThreadForSkeletonization.ACTION_START);
	    SetButtonsStartedState();
    	}
	// Stop algorithm
	else if (target == stopButton) { 
	    stopButton.disable();
	    algorithmThread.Stop();
	}
	// Initizalize principal curve 
	else if (target == initButton) { 
	    onlyInitialize = true;
	    SetButtonsStartedState();
	    algorithmThread.SetAction(PrincipalCurveAlgorithmThreadForSkeletonization.ACTION_INIT);
	}
	// Fitting and smoothing
	else if (target == fitAndSmoothButton) { 
	    onlyFitAndSmooth = true;
	    SetButtonsStartedState();
	    algorithmThread.SetAction(PrincipalCurveAlgorithmThreadForSkeletonization.ACTION_FIT_AND_SMOOTH);
	}
	// Restructuring
	else if (target == restructureButton) { 
	    onlyRestructure = true;
	    SetButtonsStartedState();
	    algorithmThread.SetAction(PrincipalCurveAlgorithmThreadForSkeletonization.ACTION_RESTRUCTURE);
	}
	// Setting data colors, sizes, and types
	else if (target == colorsAndSizesMenuItem) { 
	    Dialog dialog = canvas.GetCustomizingDialog(this);
	    dialog.show();
	}
	// Pack center canvas
	else if (target == packMenuItem) { 
	    canvas.Pack();	
	    pack();
	}
	// Quit
	else if (target == quitMenuItem) {
	    postEvent(new Event(this,Event.WINDOW_DESTROY,""));
	}
	// Select built in sample
	else {
	    for(int i=0;i<10;i++) {
		for(int j=0;j<10;j++) {
		    for(int k=0;k<10;k++) {
			if (target == digitsMenuItems[i][j][k]) {
			    DownLoad(URL_SAMPLE_DIRECTORY + "Number" + i + "/n0" + j + "" + k + ".bm");
			}
		    }
		}
	    }
	}
	return true;
    }

    public boolean handleEvent(Event event) {
	//	System.out.println(event);
	// Quit
	if (event.id == Event.WINDOW_DESTROY) {
	    if (Utilities.Environment.inApplet) {
		dispose();
	    } else {
		System.exit(0);
	    }
	}
	//Debug event
	else if (event.target == diagnosisTextField && event.arg != null) {
	    algorithmThread.ReleaseStep();
	    if (event.id == DebugEvent.FINISHED_DEBUG_THREAD) {
		if (event.arg.equals(algorithmThread.DEBUG_INIT) ){
  		    canvas.repaint();
		    if (onlyInitialize) {
			SetButtons();
			onlyInitialize = false;
		    }
		}
		else if (event.arg.equals(algorithmThread.DEBUG_FIT_AND_SMOOTH)) {
  		    canvas.repaint();
		    if (onlyFitAndSmooth) {
			SetButtons();
			onlyFitAndSmooth = false;
		    }
		}
		else if (event.arg.equals(algorithmThread.DEBUG_RESTRUCTURE)) {
  		    canvas.repaint();
		    if (onlyRestructure) {
			SetButtons();
			onlyRestructure = false;
		    }
		}
		else if (event.arg.equals(algorithmThread.DEBUG_ALL)) {
  		    canvas.repaint();
		    SetButtons();
		}
	    }
	    else if (event.id == DebugEvent.STARTED_DEBUG_THREAD) {
		if (event.arg.equals(algorithmThread.DEBUG_FIT_AND_SMOOTH)) {
		    stopButton.enable();
		}
		else if (event.arg.equals(algorithmThread.DEBUG_RESTRUCTURE)) {
		    stopButton.disable();
		}
	    }
	    else if (event.id == DebugEvent.ITERATE_DEBUG_THREAD) {
		if (event.arg.equals(algorithmThread.DEBUG_FIT_AND_SMOOTH)) {
  		    canvas.repaint();
		}
	    }
	}

        return super.handleEvent(event);    
    }

    private boolean LoadData(int index,String debugString,String fileName) {
	diagnosisTextField.setText(debugString);
	File f = new File(fileName);
	try {
	    loadableData[index].Load(f);
	}
 	catch (Throwable exception) {
	    diagnosisTextField.setText("");
	    return false;
	}
	diagnosisTextField.setText("");
	return true;
    }
    
    private void Load(String directory,String fileName) {
	if (fileName.equals(""))
	    templateFileName = "image.pbm";
	else
	    templateFileName = fileName;
	loadableData[0] = new Template();
	LoadData(0,"Loading template...",directory + templateFileName);
	// We make sure that there is always a valid template loaded
	if (((Template)loadableData[0]).GetXSize() >= 2 && ((Template)loadableData[0]).GetYSize() >= 2 &&
	    (sample = ((Template)loadableData[0]).ConvertToSample()).GetSize() > 2) {
	    characterTemplate = (Template)loadableData[0];
	    characterTemplate.FilterSaltAndPepperNoise();
	    canvas.SetObjectToPaint(canvas.sample,characterTemplate);	    
	    canvas.ResetBorders(characterTemplate);

	    principalCurve = new PrincipalCurveOfTemplate(sample);
	    loadableData[1] = new SetOfCurves();
	    if (LoadData(1,"Loading principal curve...",directory + pgraphFileName)) {
		principalCurve.InitializeToCurves((SetOfCurves)loadableData[1],0);
	    }
	    NewPrincipalCurve();	    
	    NewSample();
	}
    }

    private boolean DownLoadData(int index,String debugString,String fileName) {
	diagnosisTextField.setText(debugString);
	try {
	    DownLoadFile downLoadFile = new DownLoadFile(fileName);
	    loadableData[index].Load(downLoadFile.GetDataInputStream());
	}
 	catch (Throwable exception) {
	    diagnosisTextField.setText("");
	    return false;
	}
	diagnosisTextField.setText("");
	return true;
    }

    private void DownLoad(String urlString) {
	String fileName = "";
	StringTokenizer t = new StringTokenizer(urlString,"/");
	boolean first = true;
	while(t.hasMoreTokens())
	    fileName = t.nextToken();
	String directory = urlString.substring(0,urlString.indexOf(fileName));
	if (!fileName.equals(""))
	    templateFileName = fileName;
	loadableData[0] = new Template();
	DownLoadData(0,"Downloading sample...",directory + templateFileName);
	// We make sure that there is always a valid sample loaded
	if (((Template)loadableData[0]).GetXSize() >= 2 && ((Template)loadableData[0]).GetYSize() >= 2 &&
	    (sample = ((Template)loadableData[0]).ConvertToSample()).GetSize() > 2) {
	    characterTemplate = (Template)loadableData[0];
	    characterTemplate.FilterSaltAndPepperNoise();
	    canvas.SetObjectToPaint(canvas.sample,characterTemplate);	    
	    canvas.ResetBorders(characterTemplate);

	    principalCurve = new PrincipalCurveOfTemplate(sample);
	    loadableData[1] = new SetOfCurves();
	    if (LoadData(1,"Downloading principal curve...",directory + pgraphFileName)) {
		principalCurve.InitializeToCurves((SetOfCurves)loadableData[1],0);
	    }
	    NewPrincipalCurve();	    
	    NewSample();
	}
    }				

    private void NewSample() {
	SetButtons();
    }

    protected void NewPrincipalCurve() {
	SetButtons();
	canvas.SetObjectToPaint(canvas.principalCurve,principalCurve);
	canvas.SetObjectToPaint(canvas.principalCurvePoints,principalCurve);
	algorithmThread = new PrincipalCurveAlgorithmThreadForSkeletonization(characterTemplate,principalCurve);
    }

    private void SetButtons() {
	sampleMenu.enable();
	initCurveMenuItem.enable();
	if (principalCurve.GetSize() >= 2) {
	    // "Paused" state
	    startButton.enable();
	    stopButton.disable();
	    initButton.enable();
	    fitAndSmoothButton.enable();
	    restructureButton.enable();
	}
	else {
	    // "Loaded" state
	    startButton.enable();
	    stopButton.disable();
	    initButton.enable();
	    fitAndSmoothButton.disable();
	    restructureButton.disable();
	}	
    }

    private void SetButtonsStartedState() {
	timeStart = System.currentTimeMillis();
	startButton.disable();
	stopButton.disable();
	initButton.disable();
	fitAndSmoothButton.disable();
	restructureButton.disable();
	sampleMenu.disable();
	initCurveMenuItem.disable();
    }

    public void SetPosition() {
	move(0,50);
	pack();
    }

    public void Recall() {
	canvas.ResetBorders(sample);
	canvas.Repaint();	
    }

    public static void main(String args[]) {
  	Utilities.Environment.inApplet = false;	
	//  	Utilities.Environment.cRoutines = true; // comment it out if you can't compile the c functions	
	VektorDDWeighted.maxWeight = 1; // for shading the color

	SkeletonizationGUI window = new SkeletonizationGUI();	
 	window.SetPosition();
	window.show();
    }


class ParametersDialog extends Dialog {
    Panel panel;
    ScrollbarTextFieldPanel penaltyCoefficientPanel;
    ScrollbarTextFieldPanel relativeLengthPenaltyCoefficientPanel;
    ScrollbarTextFieldPanel relativeChangeInCriterionThresholdPanel;
    ScrollbarTextFieldPanel tauBranchPanel;
    ScrollbarTextFieldPanel tauLoopPanel;
    ScrollbarTextFieldPanel tauFilterPanel;
    ScrollbarTextFieldPanel epsStar3Panel;
    Checkbox changeResolutionCheckbox;
    ScrollbarTextFieldPanel changeResolutionPanel;
    Button doneButton;

    final int PENALTY_COEFFICIENT_FACTOR = 1000;
    final int RELATIVE_LENGTH_PENALTY_COEFFICIENT_FACTOR = 100;
    final int RELATIVE_CHANGE_FACTOR = 100000;
    final int TAU_FACTOR = 100;
    final int FILTER_FACTOR = 1000;

    public ParametersDialog(Frame frame) {
	super(frame,"Parameters",false);
  	setLayout(new BorderLayout());
  	doneButton = new Button("Done");
  	add("East",doneButton);
	panel = new Panel();
	panel.setLayout(new GridLayout(0,1));
	penaltyCoefficientPanel = 
	    new ScrollbarTextFieldPanel("penalty coefficient: ",
					(int)(Globals.penaltyCoefficient * PENALTY_COEFFICIENT_FACTOR),
					0,1000,true,PENALTY_COEFFICIENT_FACTOR);
	panel.add(penaltyCoefficientPanel);
	relativeLengthPenaltyCoefficientPanel = 
	    new ScrollbarTextFieldPanel("length penalty coefficient at end segments: ",
					(int)(Globals.relativeLengthPenaltyCoefficient * RELATIVE_LENGTH_PENALTY_COEFFICIENT_FACTOR),
					0,1000,true,RELATIVE_LENGTH_PENALTY_COEFFICIENT_FACTOR);
	panel.add(relativeLengthPenaltyCoefficientPanel);
	relativeChangeInCriterionThresholdPanel = 
	    new ScrollbarTextFieldPanel("optimization threshold: ",
					(int)(Globals.relativeChangeInCriterionThreshold * RELATIVE_CHANGE_FACTOR),
					0,10000,true,RELATIVE_CHANGE_FACTOR);
	panel.add(relativeChangeInCriterionThresholdPanel);
	tauBranchPanel = 
	    new ScrollbarTextFieldPanel("tau_branch: ",
					(int)(GlobalsSkeletonization.tauBranch * TAU_FACTOR),
					0,1000,true,TAU_FACTOR);
	panel.add(tauBranchPanel);
	tauLoopPanel = 
	    new ScrollbarTextFieldPanel("tau_loop: ",
					(int)(GlobalsSkeletonization.tauLoop * TAU_FACTOR),
					0,1000,true,TAU_FACTOR);
	panel.add(tauLoopPanel);
	tauFilterPanel = 
	    new ScrollbarTextFieldPanel("tau_filter: ",
					(int)(GlobalsSkeletonization.tauFilter * FILTER_FACTOR),
					0,10000,true,FILTER_FACTOR);
	panel.add(tauFilterPanel);
	epsStar3Panel = 
	    new ScrollbarTextFieldPanel("eps_star3: ",
					(int)(GlobalsSkeletonization.epsStar3 * TAU_FACTOR),
					0,1000,true,TAU_FACTOR);
	panel.add(epsStar3Panel);
	changeResolutionCheckbox = new Checkbox("Change resolution before initialization",GlobalsSkeletonization.changeResolution);
	panel.add(changeResolutionCheckbox);
	changeResolutionPanel = 
	    new ScrollbarTextFieldPanel("X resolution: ",
					GlobalsSkeletonization.changeResolutionX,
					0,1000,true,1);
	panel.add(changeResolutionPanel);
	changeResolutionPanel.disable();
	add("West",panel);
	validate();
	pack();
    }    

    final public boolean action(Event event, Object arg) {
	// Done
	if (event.target == doneButton) {
	    postEvent(new Event(this,Event.WINDOW_DESTROY,""));
	}
	return true;
    }

    final public boolean handleEvent(Event event) {
	// Done
	if (event.id == Event.WINDOW_DESTROY) {
	    dispose();
	}
	else if (event.target == penaltyCoefficientPanel && ScrollbarTextFieldPanel.ScrollbarEvent(event)) {
	    Globals.penaltyCoefficient = ((Integer)event.arg).intValue()/(double)PENALTY_COEFFICIENT_FACTOR;
	}
	else if (event.target == relativeLengthPenaltyCoefficientPanel && ScrollbarTextFieldPanel.ScrollbarEvent(event)) {
	    Globals.relativeLengthPenaltyCoefficient = ((Integer)event.arg).intValue()/(double)RELATIVE_LENGTH_PENALTY_COEFFICIENT_FACTOR;
	}
	else if (event.target == relativeChangeInCriterionThresholdPanel && ScrollbarTextFieldPanel.ScrollbarEvent(event)) {
	    Globals.relativeChangeInCriterionThreshold = ((Integer)event.arg).intValue()/(double)RELATIVE_CHANGE_FACTOR;
	}
	else if (event.target == tauBranchPanel && ScrollbarTextFieldPanel.ScrollbarEvent(event)) {
	    GlobalsSkeletonization.tauBranch = ((Integer)event.arg).intValue()/(double)TAU_FACTOR;
	}
	else if (event.target == tauLoopPanel && ScrollbarTextFieldPanel.ScrollbarEvent(event)) {
	    GlobalsSkeletonization.tauLoop = ((Integer)event.arg).intValue()/(double)TAU_FACTOR;
	}
	else if (event.target == tauFilterPanel && ScrollbarTextFieldPanel.ScrollbarEvent(event)) {
	    GlobalsSkeletonization.tauFilter = ((Integer)event.arg).intValue()/(double)FILTER_FACTOR;
	}
	else if (event.target == epsStar3Panel && ScrollbarTextFieldPanel.ScrollbarEvent(event)) {
	    GlobalsSkeletonization.epsStar3 = ((Integer)event.arg).intValue()/(double)TAU_FACTOR;
	}
	else if (event.target == changeResolutionCheckbox) {
	    GlobalsSkeletonization.changeResolution = changeResolutionCheckbox.getState();
	    if (GlobalsSkeletonization.changeResolution)
		changeResolutionPanel.enable();
	    else
		changeResolutionPanel.disable();	    
	}
	else if (event.target == changeResolutionPanel && ScrollbarTextFieldPanel.ScrollbarEvent(event)) {
	    GlobalsSkeletonization.changeResolutionX = ((Integer)event.arg).intValue();
	}
        return super.handleEvent(event);    
    }
}

class InitCurveDialog extends Dialog {
    SkeletonizationGUI parent;
    Panel southPanel;
    Button doneButton;
    SkeletonizationCanvas canvas;
    Window balloon;
    TextField balloonTextField;
    Checkbox balloonCheckbox;
    

    SetOfCurves initialCurves;
    boolean startedDrawingInitialCurves;

    public InitCurveDialog(SkeletonizationGUI frame,SkeletonizationCanvas in_canvas) {
	super(frame,"Initialize principal curve",true);
	parent = frame;
      	setLayout(new BorderLayout());
	southPanel = new Panel();
	southPanel.setLayout(new GridLayout(1,0));
	doneButton = new Button("Done");
	southPanel.add(doneButton);
      	balloonCheckbox = new Checkbox("Coordinate balloon");
      	balloonCheckbox.setState(false);
      	southPanel.add(balloonCheckbox);
	add("South",southPanel);
	canvas = in_canvas;
	add("Center",canvas);

	balloonTextField = new TextField(canvas.ReConvert(0,0).toString());
	balloonTextField.setEditable(false);
	balloon = new Window(parent);
	balloon.add(balloonTextField);
	balloon.setBackground(ColorChoice.GetColor("Yellow"));
	balloonTextField.setBackground(ColorChoice.GetColor("Yellow"));
	balloon.validate();
	balloon.pack();
	
	initialCurves = parent.principalCurve.ConvertToCurves();
	if (initialCurves == null)
	    initialCurves = new SetOfCurves();
	startedDrawingInitialCurves = true;
	initialCurves.StartNewCurve();
	initialCurves.AddPoint(canvas.ReConvert(0,0));

	parent.principalCurve.Reset();
	canvas.SetObjectToPaint(canvas.principalCurve,initialCurves);
	canvas.SetObjectToPaint(canvas.principalCurvePoints,initialCurves);

	validate();
	pack();
    }    

    final public boolean action(Event event, Object arg) {
	// Done
	if (event.target == doneButton) {
	    parent.canvasPanel.add(canvas);
	    canvas.Pack();	
	    parent.pack();
  	    if (initialCurves.Valid()) {
		((PrincipalCurveClass)parent.principalCurve).
		    InitializeToCurves(initialCurves,2.001 * canvas.GetCanvasSize()/canvas.getSize().width); // two pixels
	    }
	    parent.NewPrincipalCurve();
	    postEvent(new Event(this,Event.WINDOW_DESTROY,""));
	}
	else if (event.target == balloonCheckbox) {
	    if (balloonCheckbox.getState())
		balloon.show();		
	    else
		balloon.hide();		
	}
	return true;
    }

    final public boolean handleEvent(Event event) {
	// Done
	if (event.id == Event.WINDOW_DESTROY) {	    
	    balloon.dispose();
	    dispose();
	}
	return super.handleEvent(event);    
    }

    public boolean mouseDown(Event event, int x, int y) {
	Point p = canvas.location();
	x -= p.x;
	y -= p.y;
	if (canvas.inside(x,y)) {
	    if (parent.logFile != null && Utilities.Environment.inApplet) {
		parent.logFile.AppendString((((new Date()).getTime() - parent.zeroTime)/1000.0) + ":\t" + 
					    Utilities.Misc.MouseButton(event.modifiers) + " MouseDown " + x + " " + y  + "\n");
	    }	    
	    // left button
	    if (event.modifiers == 0) {
		if (!startedDrawingInitialCurves) {
		    initialCurves.Reset();
		    startedDrawingInitialCurves = true;
		    initialCurves.AddPoint(canvas.ReConvert(x,y));
		}
		doneButton.disable();
		initialCurves.AddPoint(canvas.ReConvert(x,y));
	    }
	    // right button
	    else if (event.modifiers == 4) {
		startedDrawingInitialCurves = false;		
		if (initialCurves.Valid())
		    doneButton.enable();
	    }
	    // middle button
	    else if (event.modifiers == 8) {
		initialCurves.StartNewCurve();
		initialCurves.AddPoint(canvas.ReConvert(x,y));
	    }
	    canvas.SetObjectToPaint(canvas.principalCurve,initialCurves);
	    canvas.SetObjectToPaint(canvas.principalCurvePoints,initialCurves);
	}
	return true;
    }

    public boolean mouseMove(Event event, int x, int y) {
	balloon.setLocation(location().x+x+5,location().y+y+5);
	Point p = canvas.location();
	x -= p.x;
	y -= p.y;
	if (canvas.inside(x,y)) {
	    Vektor vektor = canvas.ReConvert(x,y);
	    balloonTextField.setText(vektor.toString());
	    if (startedDrawingInitialCurves) {
		initialCurves.UpdateLastPoint(vektor);
		canvas.SetObjectToPaint(canvas.principalCurve,initialCurves);
		canvas.SetObjectToPaint(canvas.principalCurvePoints,initialCurves);
	    }
	}
	return true;
    }
}

}
    
